import { MolangVariableMap, system, Player, world, Vector, Direction } from "@minecraft/server";
import { getObjective, getScore, getMainhand, playSound, rotateVector } from "./utils";
import { Entity } from "@minecraft/server";
import { dirMap } from "./globals";

const chestBlow = {
  "ra:citori_725": {
    mul: 1.5,
    heightMul: 1,
    minLimbDist: 3,
    blowDist: 4,
    chestDist: 7,
    legDist: 7,
  },
  "ra:remington_870": {
    mul: 1.5,
    heightMul: 1,
    minLimbDist: 3,
    blowDist: 4,
    chestDist: 7,
    legDist: 7,
  },
  "ra:awm": {
    mul: 2.25,
    heightMul: 1,
    minLimbDist: 64,
    blowDist: 6,
    chestDist: 64,
    legDist: 10,
  },
  "ra:mini_gun": {
    mul: 0.9,
    heightMul: 1,
    minLimbDist: 48,
    blowDist: 8,
    chestDist: 48,
    legDist: 48,
  },
};

const lastHitObj = getObjective("ra_lasthit");

/**
 *
 * @param {Entity} entity
 * @param {number} skinId
 * @param {'arm'|'leg'} limbType
 * @param {'l'|'r'|undefined} side
 */
function detachLimb(entity, skinId, limbType, side) {
  if (entity.hasTag(`no_${limbType}.${side}`)) return;

  if (!side) side = Math.random() > 0.5 ? "r" : "l";
  entity.triggerEvent(`ra:remove_${limbType}.${side}`);

  const entLoc = entity.location;

  const isCrawler = entity.getComponent("is_stackable");
  let limbHeight = limbType == "arm" ? (isCrawler ? 0.55 : 1.5) : 0;
  let limbPos;
  if (side == "r") {
    limbPos = Vector.add(entLoc, { x: -0.25, y: limbHeight, z: 0 });
  } else {
    limbPos = Vector.add(entLoc, { x: 0.25, y: limbHeight, z: 0 });
  }

  entity.addTag(`no_${limbType}.${side}`);

  const limb = entity.dimension.spawnEntity(`ra:zombie_limb<ra:skin_${skinId}>`, limbPos);
  limb.setRotation(0, Math.random()*360)
  limb.triggerEvent("ra:despawn_timer");
  const angle = side == "r" ? -135 : 135;
  limb.applyImpulse(
    Vector.add(Vector.multiply(rotateVector(entity.getViewDirection(), angle), 0.4), { x: 0, y: 0.05, z: 0 })
  );
}

/**
 *
 * @param {*} blowConf
 * @param {Entity} entity
 * @param {number} skinId
 * @param {import("@minecraft/server").Vector3} hitVector
 * @param {'r'|'l'|undefined} sideChar
 */
function blowback(blowConf, dist, entity, skinId, hitVector, sideChar) {
  const entLoc = entity.location;
  const dim = entity.dimension;

  const isCrawler = entity.getComponent("minecraft:is_stackable");
  const chestHeight = isCrawler ? 0.275 : 1;
  const chestLoc = Vector.add(entLoc, { x: 0, y: chestHeight, z: 0 });
  const backLoc = Vector.subtract(chestLoc, Vector.multiply(entity.getViewDirection(), 0.5));

  if (dist <= blowConf.chestDist) {
    if (dist <= blowConf.blowDist) {
      const hitDir = Vector.add(Vector.multiply(hitVector, blowConf.mul * 0.4), {
        x: 0,
        y: 0.23 * blowConf.heightMul,
        z: 0,
      });

      const distMul = 1 - Math.pow(dist / blowConf.blowDist, 1.5);

      if (dist >= blowConf.minLimbDist) {
        const rlimb = Math.random();
        if (rlimb >= 0.5) detachLimb(entity, skinId, "arm", sideChar);
      }

      const yAng = (Math.atan2(hitVector.z, hitVector.x) * 180) / Math.PI + 90;
      entity.teleport(entLoc, dim, 0, yAng);

      const blowback = dim.spawnEntity("ra:zombie_blowback", entLoc);
      entity.runCommand("ride @s start_riding @e[type=ra:zombie_blowback,c=1,tag=!init]");
      blowback.addTag("init");
      blowback.ownerId = entity.id;

      blowback.applyImpulse(Vector.multiply(hitDir, distMul * 2.5));
      entity.triggerEvent("ra:blowback.fly");
    }
    entity.triggerEvent("ra:remove_chest");

    playSound(dim, "splat_impact", { pitch: Math.random() * 0.75 + 0.75, volume: 3 }, { maxDistance: 60 }, entLoc);
  
    const vars = new MolangVariableMap();
    dim.spawnParticle("ra:bullet.impact_zombie", chestLoc, vars);
    dim.spawnParticle("ra:blood", chestLoc, vars);
    dim.spawnParticle("ra:zombie_head.explode_blood", chestLoc, vars);
    dim.spawnParticle("ra:zombie_head.explode_blood_1", chestLoc, vars);
    dim.spawnParticle("ra:zombie_head.explode_blood", backLoc, vars);
    dim.spawnParticle("ra:zombie_head.explode_blood_1", backLoc, vars);

    entity.kill();
  }
}

const rotMap = {
  [Direction.east]: -90,
  [Direction.west]: 90,
  [Direction.north]: 180,
  [Direction.south]: 0,
};

world.events.projectileHit.subscribe((ev) => {
  const proj = ev.projectile;
  if (proj.typeId != "ra:zombie_blowback") return;

  const blockHit = ev.getBlockHit();
  if (!blockHit) return;

  const dim = ev.dimension;
  const zomb = dim.getEntitiesAtBlockLocation(proj.location).find((v) => v.id == proj.ownerId);
  if (!zomb) return console.warn("no zombie");

  const { block, face } = blockHit;
  const rot = rotMap[face];

  let alignY = false;
  zomb.runCommandAsync("ride @s stop_riding").then(() => {
    let { x, y, z } = proj.location;
    if (alignY) y = Math.floor(y);
    zomb.teleport({ x, y, z }, dim, 0, rot ?? zomb.getRotation().y);
    // zomb.applyImpulse(proj.getVelocity());
    proj.triggerEvent("ra:despawn");
  });

  if (rot !== undefined) {
    const above = dim.getBlock(Vector.add(block, { x: 0, y: 1, z: 0 }));
    if (above.isSolid()) {
      zomb.triggerEvent("ra:blowback.wall");

      const vars = new MolangVariableMap();
      const { x, y, z } = dirMap[face];
      const dirVec = new Vector(x, y, z);
      vars.setVector3("variable.block_face", dirVec);
      const partLoc = Vector.multiply(Vector.add(ev.location, Vector.multiply(dirVec, 0.01)), { x: 1, y: 0, z: 1 });

      let prevY = zomb.location.y + 0.02;
      let startTick = system.currentTick;
      const int = system.runInterval(() => {
        const newY = zomb.location.y;
        if (Math.abs(prevY - newY) < 0.005 && system.currentTick - startTick > 5) {
          system.clearRun(int);
          return;
        }
        prevY = newY;
        const newPartLoc = Vector.add(partLoc, { x: 0, y: newY + 0.75, z: 0 });
        dim.spawnParticle("ra:blood_splat.wall.3", newPartLoc, vars);
        dim.spawnParticle("ra:blood_splat.wall.3", Vector.add(newPartLoc, { x: 0, y: 0.4, z: 0 }), vars);
      }, 1);
      system.runTimeout(() => system.clearRun(int), 20 * 2);

      const currPartLoc = Vector.add(partLoc, { x: 0, y: prevY + 1, z: 0 });
      dim.spawnParticle("ra:blood_splat.wall.2", currPartLoc, vars);
    } else {
      zomb.triggerEvent("ra:blowback.fold");
      zomb.triggerEvent("ra:no_gravity");
      alignY = true;
    }
  } else {
    const { x, y, z } = zomb.location;
    const viewDir = zomb.getViewDirection();
    const { x: vx, z: vz } = Vector.multiply(viewDir, { x: 1, y: 0, z: 1 }).normalized();

    const behind = dim.getBlock({ x: Math.floor(x - vx), y, z: Math.floor(z - vz) });

    if (behind.isSolid()) {
      zomb.triggerEvent("ra:blowback.floor_wall");
    } else {
      zomb.triggerEvent("ra:blowback.land");
    }
  }
});

world.events.projectileHit.subscribe((ev) => {
  const src = ev.source;
  if (!(src instanceof Player)) return;

  const proj = ev.projectile;
  if (proj.typeId != "ra:bullet.projectile") return;

  const entityHit = ev.getEntityHit();
  if (!entityHit) return;

  const entity = entityHit.entity;

  if (entity.typeId !== "ra:zombie") return;

  const lastHit = getScore(lastHitObj, entity.scoreboard);
  if (lastHit) {
    const dif = system.currentTick - lastHit;
    if (dif < 1) return;
  }
  entity.runCommand(`scoreboard players set @s ${lastHitObj.id} ${system.currentTick}`);

  const projLoc = ev.location;
  const entLoc = entity.location;
  const relLoc = Vector.subtract(projLoc, entLoc);
  const { x, y, z } = relLoc;

  const horPos = rotateVector(relLoc, entity.getRotation().y);
  const isRightHit = horPos.x < 0;
  const sideChar = isRightHit ? "r" : "l";

  const dim = ev.dimension;

  const gun = getMainhand(src);
  const gunId = gun.typeId ?? "minecraft:air";

  /** @type {number} */
  const skinId = entity.getComponent("minecraft:skin_id")?.value ?? 0;

  const blowConf = chestBlow[gunId];
  const dist = Vector.subtract(entLoc, src.location).length();

  const isCrawler = entity.hasComponent("is_stackable");
  const headHeight = isCrawler ? 1 : 1.5;
  const chestHeight = isCrawler ? 0.285 : 0.8;
  // world.sendMessage(`Relative: ${blowConf.chestDist}`);

  if (y > headHeight) {
    // Implement head shot stuff?
  } else if (y < chestHeight) {
    if (blowConf && dist < blowConf.legDist) {
      if (!entity.hasTag(`no_leg.${sideChar}`)) {
        detachLimb(entity, skinId, "leg", sideChar);
        playSound(dim, "splat_impact", { pitch: Math.random() * 0.75 + 1.75, volume: 3 }, { maxDistance: 60 }, entLoc);

        if (!entity.hasComponent("is_stunned") && !isCrawler) entity.triggerEvent("ra:zombie.crawler");

        // if (entity.hasComponent("is_stunned")) entity.triggerEvent("ra:remove_chest");
      }
    }
  } else {
    if (blowConf && blowConf.blowDist) {
      blowback(blowConf, dist, entity, skinId, ev.hitVector, sideChar);
    }
    if (gunId == "ra:mini_gun" && !entity.hasComponent("is_stunned")) {
      const viewVec = entity.getViewDirection();

      detachLimb(entity, skinId, "arm", sideChar);
      blowback(blowConf, dist, entity, skinId, ev.hitVector, sideChar);

      if (Math.random() > 0.5) {
        entity.triggerEvent("ra:remove_head");
        const headOff = isCrawler ? 0.5 : 1.63;
        const head = dim.spawnEntity(
          `ra:zombie_head<ra:skin_${skinId}>`,
          Vector.add(entLoc, { x: 0, y: headOff, z: 0 })
        );
        head.triggerEvent("ra:despawn_timer");

        head.applyImpulse(Vector.add(Vector.multiply(viewVec, -0.25), { x: 0, y: 0.25, z: 0 }));
      }
      entity.triggerEvent("ra:death.mini_gun");
      entity.triggerEvent("ra:remove_chest");

      playSound(dim, "splat_impact", { pitch: Math.random() * 0.75 + 0.75, volume: 3 }, { maxDistance: 60 }, entLoc);
      entity.triggerEvent("ra:death");
      entity.kill();
    }
  }

  // dim.spawnParticle("minecraft:endrod", projLoc, new MolangVariableMap());
  // world.sendMessage(`Relative: ${x} ${y} ${z}`);
});
